/*  ____    __   ____  ______
 * |    \  /  ] /    T|      T
 * |  o  )/  / Y  o  ||      |
 * |   _//  /  |     |l_j  l_j
 * |  | /   \_ |  _  |  |  |
 * |  | \     ||  |  |  |  |
 * l__j  \____jl__j__j  l__j
 *
 * dump a packet capture in a line-based, parsable format
 *
 */

#include <getopt.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include "netre.h"
#include "pcap.h"
#include "stream.h"

int version(bool error) {
    fprintf(WHICHOUT(error), "pcat v.%s - %s\n\n", PACKAGE_VERSION,
            "dump a packet capture in a line-based, parsable format");
    return error ? EX_USAGE : EX_OK;
}

int usage(bool error, char *prog) {
    int retval = version(error);
    fprintf(WHICHOUT(error), "Usage: %s [-o OUTPUT] [-] [FILE ...]\n", prog);
    fprintf(WHICHOUT(error), "\t-o OUTPUT  output filename\n");
    fprintf(WHICHOUT(error), "\tFILE ...   pcap file(s)\n");
    return retval;
}

void print_payload(struct stream *s, FILE *out) {
    if (s->len == 0)
        fprintf(out, "-");
    else
        while (s->len) fprintf(out, "%02x", read_uint8(s));
}

void process_icmp(struct stream *s, char *saddr_s, char *daddr_s, uint16_t iplen, FILE *out) {
    uint8_t  type   = read_uint8(s);
    uint8_t  code   = read_uint8(s);
    uint16_t chksum = read_uint16(s);
    uint32_t q1     = read_uint16(s);
    uint32_t q2     = read_uint16(s);

    fprintf(out, "ICMP4 %s,%u %s,%u %u,%u,%u ",
            saddr_s, type,
            daddr_s, code,
            q1, q2, iplen - ICMP_HLEN);
}

void process_tcp(struct stream *s, char *saddr_s, char *daddr_s, uint16_t iplen, FILE *out) {
    uint16_t sport  = read_uint16(s);
    uint16_t dport  = read_uint16(s);
    uint32_t seq    = read_uint32(s);
    uint32_t ack    = read_uint32(s);
    uint8_t  off    = read_uint8(s);
    uint8_t  hlen   = (uint8_t)(off >> 4) * 4;
    uint8_t  opts   = hlen < TCP_HLEN ? 0 : hlen - TCP_HLEN;
    uint8_t  flags  = read_uint8(s);
    uint16_t window = read_uint16(s);
    uint16_t chksum = read_uint16(s);
    uint16_t urgent = read_uint16(s);
    char flags_s[8];
    char opts_s[opts*2+1];

    tcp_flags(flags_s, flags);
    tcp_options(opts_s, s, opts);

    if (hlen < TCP_HLEN) fprintf(out, "!");

    fprintf(out, "TCP4 %s,%u %s,%u %u,%u,%s,%s,%u ",
            saddr_s, sport,
            daddr_s, dport,
            seq, ack, flags_s, opts_s, iplen - hlen);
}

void process_udp(struct stream *s, char *saddr_s, char *daddr_s, uint16_t iplen, FILE *out) {
    uint16_t sport  = read_uint16(s);
    uint16_t dport  = read_uint16(s);
    uint16_t len    = read_uint16(s);
    uint16_t chksum = read_uint16(s);

    if (iplen != len) fprintf(out, "!");

    fprintf(out, "UDP4 %s,%u %s,%u %u ", 
            saddr_s, sport,
            daddr_s, dport,
            len - UDP_HLEN);
}

void process_ip4(struct stream *s, FILE *out) {
    uint8_t  vhl    = read_uint8(s);
    uint8_t  tos    = read_uint8(s);
    uint16_t length = read_uint16(s);
    uint16_t id     = read_uint16(s);
    uint16_t off    = read_uint16(s);
    uint8_t  ttl    = read_uint8(s);
    uint8_t  proto  = read_uint8(s);
    uint16_t chksum = read_uint16(s);
    uint32_t saddr  = read_uint32(s);
    uint32_t daddr  = read_uint32(s);

    char saddr_s[20];
    char daddr_s[20];

    ip4_addr(saddr_s, saddr);
    ip4_addr(daddr_s, daddr);

    // Ignore options
    sskip(s, IP_IHL(vhl) - 20);
    length = length - IP_IHL(vhl);

    // skip FCS?
    s->len = length;

    DDUMP_d(proto);
    switch (proto) {
        case IPPROTO_ICMP:
            process_icmp(s, saddr_s, daddr_s, length, out); break;
        case IPPROTO_TCP:
            process_tcp( s, saddr_s, daddr_s, length, out); break;
        case IPPROTO_UDP:
            process_udp( s, saddr_s, daddr_s, length, out); break;
        default:
            fprintf(out, "P%d %s %s %d,%d ", proto, saddr_s, daddr_s, id, length); break;
    }

    print_payload(s, out);
}

void process_arp(struct stream *s, FILE *out) {
    uint16_t htype  = read_uint16(s);
    uint16_t ptype  = read_uint16(s);
    uint8_t  hlen   = read_uint8(s);
    uint8_t  plen   = read_uint8(s);
    uint16_t op     = read_uint16(s);

    uint16_t sha[]  = { read_uint16(s), read_uint16(s), read_uint16(s) };
    uint32_t spa    = read_uint32(s);

    uint16_t tha[]  = { read_uint16(s), read_uint16(s), read_uint16(s) };
    uint32_t tpa    = read_uint32(s);

    char sha_s[20], spa_s[20];
    char tha_s[20], tpa_s[20];

    mac_addr(sha_s, sha);
    ip4_addr(spa_s, spa);
    mac_addr(tha_s, tha);
    ip4_addr(tpa_s, tpa);

    fprintf(out, "ARP %s,%s %s,%s ", sha_s, spa_s, tha_s, tpa_s);

    switch(op) {
        case ARP_REQ: fprintf(out, "request,0 -");  break;
        case ARP_REP: fprintf(out, "reply,0 -");    break;
        default:      fprintf(out, "%02x,0 -", op); break;
    }
}

void print_ethernet(struct stream *s, FILE *out) {
    uint8_t  saddr[6];
    uint8_t  daddr[6];
    uint16_t vlan;             // Optional
    uint16_t ethertype;

    sread(s, &saddr, sizeof(saddr));
    sread(s, &daddr, sizeof(daddr));

    ethertype = read_uint16(s);

    if (ethertype == 0x8100) { // 802.1Q tag
        vlan = read_uint16(s);  // 3 bits QoS, 1 bit Drop Eligible, 12 bits VLAN
        ethertype = read_uint16(s);
    }

    switch (ethertype) {
        case 0x0806:          // ARP
            process_arp(s, out); break;
        case 0x8870:          // Jumbo Frames
        case 0x0800:          // IPv4
            process_ip4(s, out); break;
        case 0x86DD:          // IPv6
        default:
            DDUMP_x(ethertype);
            fprintf(out, "ETHER 0x%04x", ethertype);
    }
}

void print_frame(struct pcap_file *p, struct pcap_pkthdr *hdr, char const *frame, FILE *out) {
    struct stream streambuf;
    struct stream *s = &streambuf;

    sinit(s, frame, hdr->caplen, NETWORK_ORDER);
    DDUMP_d(hdr->caplen);
    fprintf(out, "%d.%06d ", hdr->ts.tv_sec, hdr->ts.tv_usec);

    switch (p->linktype) {
        case LINKTYPE_ETHERNET:
            print_ethernet(s, out);
            break;
        case LINKTYPE_RAW:
            process_ip4(s, out);
            break;
        default:
            fprintf(out, "LINK %d", p->linktype);
    }

    fprintf(out, "\n");
}

int pcat(FILE *fp, const char *fn, FILE *out, const char *outfn) {
    struct pcap_file p;

    if (pcap_open_in(&p, fp) == -1) return EX_IOERR;

    while (true) {
        struct pcap_pkthdr hdr;
        char frame[MAXFRAME];
        DDUMP("---");

        if (pcap_read_pkthdr(&p, &hdr)     == -1) break;
        if (fread(frame, hdr.caplen, 1, fp) !=  1) break;

        print_frame(&p, &hdr, frame, out);
    }
    if (ferror(fp)) {
        perror(fn);
        clearerr(fp);
        return EX_IOERR;
    }
    if (ferror(out)) {
        perror(outfn);
        clearerr(out);
        return EX_IOERR;
    }

    return EX_OK;
}

int main(int argc, char *argv[]) {
    int   opt, retval;
    FILE        *outp = stdout;
    const char *outfn = "stdout";
    FILE         *inp = stdin;
    const char  *infn = "stdin";


    while ((opt = getopt(argc, argv, "hvo:")) != -1) {
        switch (opt) {
            case 'o': /* output file */
                outp = fopen(optarg, "wb");
                if (!outp) {
                    perror(optarg);
                    return EX_OSFILE;
                }
                outfn = optarg;
                break;
            case 'v': return version(false);
            case 'h': return usage(false, argv[0]);
            default:  return usage(true,  argv[0]);
        }
    }

    /* fast forward argv */
    argv += optind;

    do {
        if (*argv) {
            if (!strncmp(*argv, "-", 1)) {
                inp = stdin;
            } else if ((inp = fopen(*argv, "rb")) == NULL) {
                perror(*argv);
                argv++;
                continue;
            }
            infn = *argv++;
        }

        retval = pcat(inp, infn, outp, outfn);

        if (inp    != stdin) fclose(inp);
        if (retval != EX_OK) return retval;
    } while (*argv);

    if (outp != stdout) fclose(outp);

    return retval;
}
