/*
 *  _ __  _ __ ___   ___ _ __ __ _  ___
 * | '_ \| '_ ` _ \ / _ \ '__/ _` |/ _ \
 * | |_) | | | | | |  __/ | | (_| |  __/
 * | .__/|_| |_| |_|\___|_|  \__, |\___|
 * |_|                       |___/
 *
 * Merges pcap files, outputting time-ordered pcap stream
 *
 */

#include <err.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <sysexits.h>
#include <getopt.h>
#include "pcap.h"
#include "netre.h"

struct input_file {
    int                active;
    struct pcap_file   p;
    struct pcap_pkthdr next;
};

int version(bool error) {
    fprintf(WHICHOUT(error), "pmerge v.%s - %s\n\n", PACKAGE_VERSION,
            "Merges pcap files, outputting time-ordered pcap stream");
    return error ? EX_USAGE : EX_OK;
}

int usage(bool error, char *prog) {
    int retval = version(error);
    fprintf(WHICHOUT(error), "Usage: %s [-o OUTPUT] [-] [FILE ...]\n", prog);
    fprintf(WHICHOUT(error), "\t-o OUTPUT  output filename\n");
    fprintf(WHICHOUT(error), "\tFILE ...   pcap files to merge\n");
    return retval;
}

int read_next(struct input_file *file) {
    if (! file->active) return -1;

    if (pcap_read_pkthdr(&file->p, &file->next) == -1) {
        pcap_close(&file->p);
        file->active = 0;
        file->next.ts.tv_sec = 0xffffffff;
        file->next.ts.tv_usec = 0xffffffff;
        return -1;
    }

    return 0;
}

int pmerge(FILE *output, const char *outfn, struct input_file files[], int nfiles) {
    int               nopen;
    int               i;
    struct pcap_file  out;
    struct linklayer  fake = {{0xfffe,0xcade,0xadfb}, {0xfffe,0xadde,0xefbe}, 0x0008};

    if (pcap_open_out(&out, output) == -1) {
        perror("writing pcap header");
        return EX_IOERR;
    }

    nopen = nfiles;
    while (nopen) {
        struct input_file *cur = &files[0];
        char               frame[MAXFRAME];
        size_t             len;

        /* Find next oldest frame */
        for (i = 0; i < nfiles; i += 1) {
            if ( files[i].active                                      &&
               ((files[i].next.ts.tv_sec   < cur->next.ts.tv_sec)     ||
               ((files[i].next.ts.tv_sec  == cur->next.ts.tv_sec)     &&
               ( files[i].next.ts.tv_usec <  cur->next.ts.tv_usec)))) {
               cur = &files[i];
            }
        }

        /* Make sure it'll fit */
        if (cur->next.caplen > MAXFRAME) {
            warnx("huge frame (size %u)", (unsigned int)cur->next.caplen);
            return EX_SOFTWARE;
        }

        /* Read it */
        len = fread(frame, 1, cur->next.caplen, cur->p.f);
        if (len < cur->next.caplen) {
            /* Short read */
            cur->next.caplen = len;
            pcap_close(&cur->p);
            cur->active = 0;
        }

        /* Write header + frame */
        if (len) {
            if (cur->p.linktype == LINKTYPE_RAW) {
                cur->next.caplen += sizeof(fake);
                cur->next.len += sizeof(fake);
            }
            if (fwrite(&cur->next, sizeof(cur->next), 1, output) != 1) {
                perror("error");
                return EX_IOERR;
            }
            if (cur->p.linktype == LINKTYPE_RAW) {
                if (fwrite(&fake, sizeof(fake), 1, output) != 1) {
                    perror("error");
                    return EX_IOERR;
                }
            }
            if (fwrite(frame, 1, len, output) != len) {
                perror("error");
                return EX_IOERR;
            }
        }

        if (read_next(cur) == -1) nopen -= 1;
    }

    if (ferror(output)) {
        perror(outfn);
        clearerr(output);
        return EX_IOERR;
    }

    return EX_OK;
}

int main(int argc, char *argv[]) {
    int  opt, retval, nfiles = 0;
    FILE      *output = stdout;
    const char *outfn = "stdout";
    FILE          *fp = stdin;
    const char    *fn = "stdin";
    struct input_file files[argc + 1]; /* allow for unnamed stdin */

    while ((opt = getopt(argc, argv, "hvo:")) != -1) {
        switch (opt) {
            case 'o': /* output file */
                output = fopen(optarg, "wb");
                if (!output) {
                    perror("Open output");
                    return EX_OSFILE;
                }
                outfn = optarg;
                break;
            case 'v': return version(false);
            case 'h': return usage(false, argv[0]);
            default:  return usage(true,  argv[0]);
        }
    }

    /* fast foward argv */
    argv += optind;

    /* Open input files */
    do {
        struct input_file *cur = &files[nfiles];

        if (*argv) {
            if (!strncmp(*argv, "-", 1)) {
                fp = stdin;
            } else if ((fp = fopen(*argv, "rb")) == NULL) {
                perror(*argv);
                argv++;
                continue;
            }
            fn = *argv++;
        }
        DDUMP_s(fn);

        /* prime the input files */
        if (pcap_open_in(&cur->p, fp) == -1) {
            warnx("unable to process: %s", fn);
            return EX_IOERR;
        }
        cur->active = 1;

        if (read_next(cur) == 0) nfiles += 1;
    } while (*argv);

    retval = pmerge(output, outfn, files, nfiles);

    if (output != stdout) fclose(output);
    return retval;
}
