/*                    _
 *  _ __  _   _ _ __ (_) __ _
 * | '_ \| | | | '_ \| |/ _` |
 * | |_) | |_| | | | | | (_| |
 * | .__/ \__,_|_| |_|_|\__, |
 * |_|                     |_|
 *
 * filter one to many PCAPs for unique frames
 *
 */

#include <getopt.h>
#include <stdio.h>
#include <stdbool.h>
#include <sysexits.h>
#include <string.h>
#include "netre.h"
#include "pcap.h"

int version(bool error) {
    fprintf(WHICHOUT(error), "puniq v.%s - %s\n\n", PACKAGE_VERSION,
            "filter one to many PCAPs for unique frames");
    return error ? EX_USAGE : EX_OK;
}

int usage(bool error, char *prog) {
    int retval = version(error);
    fprintf(WHICHOUT(error), "Usage: %s [-o OUTPUT] [-] [FILE ...]\n", prog);
    fprintf(WHICHOUT(error), "\t-o OUTPUT  output filename\n");
    fprintf(WHICHOUT(error), "\tFILE ...   pcap files to uniq\n");
    return retval;
}

int puniq(FILE *output, const char *outfn, char *inputs[]) {
    struct pcap_pkthdr   hdr[2];
    char               frame[2][MAXFRAME];
    struct linklayer    fake[ ] = {
        {{0xfffe,0xcade,0xadfb}, {0xfffe,0x3301,0x2370}, 0x0008},
        {{0xfffe,0x3301,0x2370}, {0xfffe,0xcade,0xadfb}, 0x0008}
    };
    FILE       *inp = stdin;
    const char *inf = "stdin";
    struct pcap_file   out;
    int                cur;

    do {
        struct pcap_file p;

        if (*inputs) {
            if (!strncmp(*inputs, "-", 1)) {
                inp = stdin;
            } else if ((inp = fopen(*inputs, "rb")) == NULL) {
                perror(*inputs);
                inputs++;
                continue;
            }
            inf = *inputs++;
        }

        if (pcap_open_in(&p, inp) == -1) return EX_IOERR;

        if (!out.f) {
            if (pcap_open_out(&out, output) == -1) {
                perror("writing header");
                return EX_IOERR;
            }
        }

        cur = 0;
        while (true) {
            size_t len;

            memset(&hdr[cur], 0, sizeof(hdr[cur]));
            memset(&frame[cur], 0, MAXFRAME);

            if (pcap_read_pkthdr(&p, &hdr[cur]) == -1) break;
            len = fread(frame[cur], 1, hdr[cur].caplen, p.f);
            if (len != hdr[cur].caplen) {
                hdr[cur].caplen = len;
            }

            /* fixup the header for LINKTYPE_RAW */
            if (p.linktype == LINKTYPE_RAW) {
                hdr[cur].caplen += sizeof(fake[cur]);
                hdr[cur].len += sizeof(fake[cur]);
            }

            if (len) {
                if ((memcmp(&hdr[0], &hdr[1], sizeof(hdr[cur])) ==  0)  &&
                   (memcmp(frame[0], frame[1], len) ==  0)) {
                    /* Skip this duplicate */
                    DDUMP("duplicate!");
                } else {
                    /* write the header */
                    if (fwrite(&hdr[cur], sizeof(hdr[cur]), 1, out.f) != 1) break;
                    /* write a fake link layer if needed */
                    if (p.linktype == LINKTYPE_RAW)
                        if (fwrite(&fake[cur], sizeof(fake[cur]), 1, out.f) != 1) break;
                    /* write the frame */
                    if (fwrite(&frame[cur], 1, len, out.f) != len) break;
                }
            }

            cur = (1 - cur);
        }

        if (ferror(inp)) {
            perror(inf);
            clearerr(inp);
            return EX_IOERR;
        }
        if (inp != stdin) fclose(inp);

        if (ferror(output)) {
            perror(outfn);
            clearerr(output);
            return EX_IOERR;
        }
    } while (*inputs);

    if (output != stdout) fclose(output);
    return EX_OK;
}

int main (int argc, char *argv[]) {
    int            opt;
    FILE       *output = stdout;
    const char  *outfn = "stdout";

    while ((opt = getopt(argc, argv, "hvo:")) != -1) {
        switch (opt) {
            case 'o': /* output file */
                output = fopen(optarg, "wb");
                if (!output) {
                    perror("opening output");
                    return EX_OSFILE;
                }
                outfn = optarg;
                break;
            case 'v': return version(false);
            case 'h': return usage(false, argv[0]);
            default:  return usage(true,  argv[0]);
        }
    }

    /* fast forward argv */
    argv += optind;

    return puniq(output, outfn, argv);
}
